/* tunelfso.c -- change some relocations of ELF .so to R_386_RELATIVE,
   to reduce runtime symbol searches during exec(), dlopen(), and PLT.
   $Id: tunelfso.c,v 1.2 1999/09/24 21:00:33 jreiser Exp $

   Copyright 1999 John F. Reiser.  All rights reserved.
   Contact: jreiser@BitWagon.com
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License, version 2.
*/
#include <elf.h>
#include <sys/stat.h>  /* fstat */
#include <sys/mman.h>  /* mmap */
#include <fcntl.h>  /* open */
#include <stdio.h>  /* perror */
#include <stdlib.h>  /* malloc */
#include <string.h>  /* strncmp */
#include <unistd.h>  /* read */

static int
Elf32_good_header(Elf32_Ehdr const *const ehdr)
{
	if (0!=strncmp((void *)ehdr, ELFMAG, SELFMAG)
	||  ELFCLASS32  != ehdr->e_ident[EI_CLASS]
	||  ELFDATA2LSB != ehdr->e_ident[EI_DATA]
	||  EV_CURRENT != ehdr->e_ident[EI_VERSION]
	|| (EM_386 != ehdr->e_machine && EM_486 != ehdr->e_machine)
	|| (ET_EXEC != ehdr->e_type && ET_DYN != ehdr->e_type) ) {
		return 0;
	}
	return 1;
}

static char *
getfile(
	char const *const fname,
	size_t *p_size
)
{
	char *retval;
	struct stat statbuf;
	int const fd = open(fname, O_RDONLY);

	if (0 > fd) {
		goto fail;
	}
	if (0 > fstat(fd, &statbuf)) {
		goto fail;
	}
	*p_size = statbuf.st_size;
	retval = (char *)mmap(0, statbuf.st_size, PROT_READ|PROT_WRITE,
		MAP_PRIVATE, fd, 0);
	if (((caddr_t)-1) == (caddr_t)retval) {
		int n_read;
		retval = (char *)mmap(0, statbuf.st_size, PROT_READ|PROT_WRITE,
			MAP_ANONYMOUS, 0, 0);
		if (((caddr_t)-1) == (caddr_t)retval) {
			goto fail;
		}
		n_read = read(fd, (caddr_t)retval, statbuf.st_size);
		if (n_read != statbuf.st_size) {
			goto fail;
		}
	}
	close(fd);
	return retval;
fail:
	perror(fname);
	exit(1);
	return 0;
}

static Elf32_Ehdr *
mmap_elf(
	char const *const elf_name,
	size_t *p_elf_size
)
{
	Elf32_Ehdr *const ehdr = (Elf32_Ehdr *)getfile(elf_name, p_elf_size);
	if (0==Elf32_good_header(ehdr)) {
		perror(elf_name);
		exit(1);
	}
	return ehdr;
}

static Elf32_Shdr const *
find_section_type(
	int key,
	Elf32_Ehdr const *const ehdr
)
{
	int j = ehdr->e_shnum;
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (int)ehdr);
	for (; --j>=0; ++shdr) {
		if (shdr->sh_type == key) {
			return shdr;
		}
	}
	return 0;
}

static unsigned
get_dynamic(
	int const key,
	Elf32_Dyn const *dynp
)
{
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			return dynp->d_un.d_val;
		}
	}
	return ~0;
}

static void const *
find_dynamic(
	int const key,
	Elf32_Dyn const *dynp,
	Elf32_Ehdr const *const ehdr
)
{
	for (; DT_NULL!=dynp->d_tag; ++dynp) {
		if (dynp->d_tag == key) {
			return (void const *)(dynp->d_un.d_val + (int)ehdr);
		}
	}
	return 0;
}

static unsigned
find_offset(
	unsigned const offset,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ehdr->e_phoff + (int)ehdr);
	int j = ehdr->e_phnum;
	for (; --j>=0; ++phdr) if (PT_LOAD == phdr->p_type) {
		unsigned const t = offset - phdr->p_vaddr;
		if (t < phdr->p_filesz) {
			return t + phdr->p_offset;
		}
	}
	return 0;
}

static unsigned
find_value(
	Elf32_Sym const *const sp,
	Elf32_Ehdr const *const ehdr
)
{
	Elf32_Shdr const *shdr = (Elf32_Shdr const *)(ehdr->e_shoff + (int)ehdr);
	if (STB_LOCAL == ELF32_ST_BIND(sp->st_info))
	if (sp->st_shndx < ehdr->e_shnum) {
		return sp->st_value + find_offset(shdr[sp->st_shndx].sh_addr, ehdr);
	}
	return sp->st_value;
}

#include <regex.h>

static int
matches(
	char const *const name,
	regex_t const *pat  /* array of patterns */
)
{
	for (; 0!=pat->buffer; ++pat) {
		if (0==regexec(pat, name, 0, 0, 0)) {
			return 1;
		}
	}
	return 0;
}

static int
partial_rel1(
	Elf32_Rel *rp,
	unsigned const rsz,
	Elf32_Sym const *const symtab,
	char const *const strtab,
	Elf32_Ehdr const *const ehdr,
	regex_t const *const ipat,
	regex_t const *const xpat
)
{
	Elf32_Rel *const rp_end = (Elf32_Rel *)(rsz + (char *)rp);
	int n = 0;
	for (; rp < rp_end; ++rp) {
		Elf32_Sym const *const sp = ELF32_R_SYM(rp->r_info) + symtab;
		char const *const name = sp->st_name + strtab;
		if ( (  R_386_GLOB_DAT == ELF32_R_TYPE(rp->r_info)
		     || R_386_32       == ELF32_R_TYPE(rp->r_info) )
		&& SHN_UNDEF  != sp->st_shndx
		&& SHN_ABS    != sp->st_shndx
		&& SHN_COMMON != sp->st_shndx
		&&  matches(name, ipat)
		&& !matches(name, xpat) ) {
			fprintf(stderr,"%s\n",name);
			*(unsigned *)(find_offset(rp->r_offset, ehdr) + (int)ehdr) += find_value(sp, ehdr);
			rp->r_info = ELF32_R_INFO(0, R_386_RELATIVE);
			++n;
		}
	}
	return n;
}

static int
partial_rel2(
	Elf32_Rel *rp,
	unsigned const rsz,
	Elf32_Sym const *const symtab,
	char const *const strtab,
	Elf32_Ehdr const *const ehdr,
	regex_t const *const ipat,
	regex_t const *const xpat
)
{
	Elf32_Rel *const rp_end = (Elf32_Rel *)(rsz + (char *)rp);
	int n = 0;
	for (; rp < rp_end; ++rp) {
		Elf32_Sym const *const sp = ELF32_R_SYM(rp->r_info) + symtab;
		char const *const name = sp->st_name + strtab;
		if (R_386_JMP_SLOT == ELF32_R_TYPE(rp->r_info)
		&& SHN_UNDEF  != sp->st_shndx
		&& SHN_ABS    != sp->st_shndx
		&& SHN_COMMON != sp->st_shndx
		&&  matches(name, ipat)
		&& !matches(name, xpat) ) {
			fprintf(stderr,"%s\n",name);
			*(unsigned *)(find_offset(rp->r_offset, ehdr) + (int)ehdr) = find_value(sp, ehdr);
			++n;
		}
	}
	return n;
}

static regex_t *
getpattern(char *arg)
{
	regex_t *retval, *rp;
	char const *argend;
	int n_retval;
	size_t len;
	if ('@' == arg[0]) {
		arg = getfile(1+arg, &len);
	}
	else {
		len = strlen(arg);
	}
	argend = len + arg;

	n_retval = 5 + (len>>1);
	rp = retval = (regex_t *)malloc(n_retval * sizeof(regex_t));
	memset(retval, 0, n_retval * sizeof(regex_t));

	while (arg < argend) {
		char *ap;
		int err;
		while (isspace(*arg)) {
			++arg;
		}
		/* This implementation requires a space immediately before '#',
		   if there is a token to the left of the '#'.  Tough.
		*/
		if ('#'==*arg) {  /* skip over comment */
			while (arg < argend && '\n'!=*arg) {
			 	++arg;
			}
			continue;
		}
		ap = arg;
		while (arg < argend && !isspace(*arg)) {
			++arg;
		}
		if (0!=*arg) {  /* allow read-only single string */
			*arg = 0;
		}
		++arg;
		err = regcomp(rp, ap, REG_EXTENDED|REG_NOSUB);
		if (0!=err) {
			char errbuf[400];
			size_t esize = regerror(err, rp, errbuf, sizeof(errbuf));
			fprintf(stderr, "*** %s\n", errbuf);
			exit(1);
		}
		++rp;
		if (--n_retval <= 0) {
			fprintf(stderr, "*** Need more pattern space.\n");
			exit(1);
		}
	}
	return retval;
}

static void
badarg(char const *arg)
{
	fprintf(stderr,"*** warning: ignoring `%s'\n", arg);
}

static char const *
do_options(
	int argc,
	char *const *argv,
	char **include,
	char **exclude
)
{
	while (0 < argc) {
		char *arg = *argv;
		if ('-'==arg[0]) {
			if ('i'== arg[1]) {
				*include = 2+arg;
			}
			else if ('x'==arg[1]) {
				*exclude = 2+arg;
			}
			else {
				badarg(arg);
			}
		}
		else if (1<argc) {
			badarg(arg);
		}
		else {
			break;
		}
		--argc;
		++argv;
	}
	return *argv;
}

int
main(
	int const argc,
	char *const *const argv,
	char const *const *const envp
)
{
	int j;

	char *include = "^_";
	char *exclude = "";
	
	char const *const elf_name = do_options(-1+argc, 1+argv, &include, &exclude);
	regex_t const *const ipat = getpattern(include);
	regex_t const *const xpat = getpattern(exclude);
	size_t elf_size;
	Elf32_Ehdr const *const ehdr = mmap_elf(elf_name, &elf_size);
	Elf32_Shdr const *const dynsec = (Elf32_Shdr const *)find_section_type(SHT_DYNAMIC, ehdr);
	Elf32_Dyn const *const dynamic = (Elf32_Dyn const *)(dynsec->sh_offset + (int)ehdr);
	unsigned const *const hashtab = (unsigned const *)find_dynamic(DT_HASH, dynamic, ehdr);
	int const n_bucket = hashtab[0];
	int const n_chain = hashtab[1];  /* == n_symtab */
	unsigned const *const bucket = &hashtab[2];
	unsigned const *const chain = n_bucket + bucket;

	Elf32_Sym const *const symtab = (Elf32_Sym const *)find_dynamic(DT_SYMTAB, dynamic, ehdr);

	char const *const strtab = (char const *)find_dynamic(DT_STRTAB, dynamic, ehdr);
	unsigned const sz_strtab = get_dynamic(DT_STRSZ, dynamic);

	Elf32_Rel *const rel1 = (Elf32_Rel *)find_dynamic(DT_REL, dynamic, ehdr);
	unsigned const sz_rel1 = get_dynamic(DT_RELSZ, dynamic);

	Elf32_Rel *const rel2 = (Elf32_Rel *)find_dynamic(DT_JMPREL, dynamic, ehdr);
	unsigned const sz_rel2 = get_dynamic(DT_PLTRELSZ, dynamic);

	partial_rel1(rel1, sz_rel1, symtab, strtab, ehdr, ipat, xpat);
	partial_rel2(rel2, sz_rel2, symtab, strtab, ehdr, ipat, xpat);

	if (elf_size != write(1, ehdr, elf_size)) {
		perror("write");
		exit(1);
	}
	return 0;
}
